Attach("HypergeometricEulerFactorKNew.m");

basechange := function(factor, f);
  _<T> := Parent(factor);
  factor := ChangeRing(factor, FieldOfFractions(BaseRing(T)));
  factor := CharacteristicPolynomial(MatrixWithGivenCharacteristicPolynomial(factor)^f);
  factor /:= Coefficient(factor,0);  // makes polynomial 1 + ...
  factor := Parent(T)!factor;
  factor := Evaluate(factor, T^f);
  return factor;
end function;

powerf := function(factor, f);
  _<T> := Parent(factor);
  factor := ChangeRing(factor, FieldOfFractions(BaseRing(T)));
  factor := CharacteristicPolynomial(MatrixWithGivenCharacteristicPolynomial(factor)^f);
  factor /:= Coefficient(factor,0);  // makes polynomial 1 + ...
  factor := Parent(T)!factor;
//  factor := Evaluate(factor, T^f);
  return factor;
end function;

K3EulerFactor := function(family, p, psi);
  t := psi^4;   // Magma uses inverse convention!!!!  AAAAUUUUGGGGHGHHHH!!
  _<T> := PolynomialRing(Integers());

  factor0 := EulerFactor(LSeries(HypergeometricData([[1/4,1/2,3/4],[0,0,0]]),t),p);
  factor0 := Parent(T)!factor0;
  // print factor0;

  if family eq "F4" then
    factor1 := EulerFactor(LSeries(HypergeometricData([[1/4,3/4],[0,1/2]]),t),p);
    factor1 := Parent(T)!factor1;
    // factor1 := Evaluate(factor1, LegendreSymbol(1-t,p)*T)^3;
    if p mod 4 eq 3 then
      factor1 := Evaluate(factor1, -T)^3;
    else
      factor1 := Evaluate(factor1, T)^3;
    end if;
    // print factor1, LegendreSymbol(1-t,p);

    factor2 := EulerFactor(LSeries(HypergeometricData([[1/2],[0]]),t),p);
    factor2 := Parent(T)!factor2;
    if p mod 4 eq 1 then
      factor2 := Evaluate(factor2, (-1)^((p-1) div 4)*T)^12;
    else
      factor2 := basechange(factor2, 2);
      factor2 := Evaluate(factor2, T);
      factor2 := Evaluate(factor2, (-1)^((p^2-1) div 4)*T)^6;
    end if;

    factor0 *:= Evaluate(factor1*factor2, p*T);

  elif family eq "F1L3" then
    if IsSquare(GF(7)!p) then
      factor1 := HypergeometricEulerFactorsKNew([[[1/14,9/14,11/14],[0,1/4,3/4]],
                                                 [[3/14,5/14,13/14],[0,1/4,3/4]]],t^-1,p : ZigZagD := 0);  // like the square
      factor1 := Parent(T)!factor1;
    else
      factor1 := HypergeometricEulerFactorKNew([1/14,9/14,11/14],[0,1/4,3/4],t^-1,p^2 : ZigZagD := 0);
      factor1 := Parent(T)!factor1;
      factor1 := Evaluate(factor1,T^2);
    end if;
    if p mod 7 in [1,6] then // IsPower(GF(7)!p,3) then 
      factor1 := Evaluate(factor1, T)^3;
    else
      factor1 := basechange(factor1, 3);
    end if;

    factor0 *:= Evaluate(factor1, p*T);

  elif family eq "F2L2" then
    if p mod 8 eq 1 then
      trivfactor := (1-T)^3;
    else
      trivfactor := (1-T)*(1+T)^2;
    end if;
    print trivfactor;

    factor1 := EulerFactor(LSeries(HypergeometricData([[1/4,3/4],[0,1/2]]),t),p);
    factor1 := Parent(T)!factor1;
    factor1 := Evaluate(factor1, (-1)^((p-1) div 2)*T);
    // print factor1;
    
    factor2 := EulerFactor(LSeries(HypergeometricData([[1/2],[0]]),t),p);
    factor2 := Parent(T)!factor2;
    if p mod 4 eq 1 then
      factor2 := Evaluate(factor2, (-1)^((p-1) div 4)*T)^2;
    else
      factor2 := basechange(factor2, 2);
      factor2 := Evaluate(factor2, (-1)^((p^2-1) div 4)*T);
    end if;
    // print factor2;

    if p mod 4 eq 1 then
      factor3 := HypergeometricEulerFactorsKNew([[[1/8,5/8],[0,1/4]],
                                                 [[3/8,7/8],[0,3/4]]],t^-1,p : ZigZagD := 0);
      factor3 := Parent(T)!factor3;
      if p mod 8 eq 1 then
        if (GF(p)!2)^((p-1) div 4) eq -1 then
          factor3 := Evaluate(factor3, -T);
        end if;
        factor3 := factor3^2;
      else
        factor3 := basechange(factor3, 2);  // inert in QQ(sqrt(-1)) to QQ(zeta_8)
      end if;
    else
      factor3 := HypergeometricEulerFactorKNew([1/8,5/8],[0,1/4],t^-1,p^2 : Precision := 10, ZigZagD := 0);
      factor3 := Evaluate(Parent(T)!factor3,T^2)^2; // inert, then split in biquadratic field QQ(zeta_8)
    end if;
    if p mod 8 ne 1 and (GF(p)!2)^((p^2-1) div 4) eq -1 then // need to replace T^2 by -T^2
      factor3cf := Coefficients(factor3);
      for i := 1 to #factor3cf do
        if i mod 2 eq 1 then // start numbering at 1 for 0th coefficient
          factor3cf[i] *:= (-1)^((i-1) div 2); 
        else
          assert factor3cf[i] eq 0;
        end if;
      end for;
      factor3 := Parent(T)!factor3cf;
    end if;
    // print factor3;
    
    factor0 *:= Evaluate(Parent(T)!(factor1*factor2*factor3*trivfactor^2), p*T);

  elif family eq "L2L2" then
    factor1 := EulerFactor(LSeries(HypergeometricData([[1/4,3/4],[0,1/2]]),t),p);
    factor1 := Parent(T)!factor1;
    factor1 := Evaluate(factor1, (-1)^((p-1) div 2)*T);

    factor2 := EulerFactor(LSeries(HypergeometricData([[1/8,3/8,5/8,7/8],[0,1/4,1/2,3/4]]),t),p);
    factor2 := Parent(T)!factor2;
    if p mod 4 eq 1 then
      factor2 := Evaluate(factor2, (-1)^((p-1) div 4)*LegendreSymbol(psi,p)*T)^2;
      trivfactor := (1-T)^8;
    else
      factor2 := basechange(factor2, 2);
      factor2 := Evaluate(factor2, (-1)^((p-1) div 4)*LegendreSymbol(psi,p)*T);  // shouldn't matter, because square
      trivfactor := (1-T)^4*(1+T)^4;
    end if;

    factor0 *:= Evaluate(factor1*factor2*trivfactor, p*T);

  elif family eq "L4" then
    factor1 := EulerFactor(LSeries(HypergeometricData([[1/5,2/5,3/5,4/5],[0,1/4,1/2,3/4]]),1/t),p);
    factor1 := Parent(T)!factor1;

    if p mod 5 eq 1 then
      factor1 := Evaluate(factor1, T)^4;
     elif p mod 5 eq 4 then
      factor1 := Evaluate(basechange(factor1, 2), T)^2;
    else 
      factor1 := Evaluate(basechange(factor1, 4), T);
    end if;
    
    trivfactor := (1-T)^2;
    
    factor0 *:= Evaluate(factor1*trivfactor, p*T);

  else
    error "Family not recognized!";
  end if;

  return factor0;
end function;



K3Trace := function(family, q, psi);
  t := psi^4;   // Magma uses inverse convention!!!!  AAAAUUUUGGGGHGHHHH!!
  _<T> := PolynomialRing(Integers());

  tr0 := HypergeometricTrace(HypergeometricData([[1/4,1/2,3/4],[0,0,0]]),t,q); 

  if family eq "F4" then
    tr1 := HypergeometricTrace(HypergeometricData([[1/4,3/4],[0,1/2]]),t,q);
    tr1 *:= (-1)^((q-1) div 2)*3;

    if q mod 4 eq 1 then
      tr2 := HypergeometricTrace(HypergeometricData([[1/2],[0]]),t,q);
      tr2 *:= (-1)^((q-1) div 4)*12;
    else
      tr2 := 0;
    end if;

    tr0 +:= q*(tr1+tr2);

  elif family eq "F1L3" then
    if q mod 7 eq 1 then
      tr1 := HypergeometricTraceKNew([1/14,9/14,11/14],[0,1/4,3/4],t^-1,q : ZigZagD := 0) + 
             HypergeometricTraceKNew([3/14,5/14,13/14],[0,1/4,3/4],t^-1,q : ZigZagD := 0);
      tr1 *:= 3;
    else
      tr1 := 0;
    end if;
    
    tr0 +:= q*tr1;

  elif family eq "F2L2" then
    if q mod 8 eq 1 then
      trivtr := 6;
    else
      trivtr := -2;
    end if;

    tr1 := HypergeometricTrace(HypergeometricData([[1/4,3/4],[0,1/2]]),t,q);
    tr1 *:= (-1)^((q-1) div 2);

    if q mod 4 eq 3 then
      tr2 := 0;
    else
      tr2 := HypergeometricTrace(HypergeometricData([[1/2],[0]]),t,q);
      tr2 *:= (-1)^((q-1) div 4)*2;
    end if;
    
    if q mod 8 eq 1 then
      tr3 := HypergeometricTraceKNew([1/8,5/8],[0,1/4],t^-1,q : ZigZagD := 0) + 
             HypergeometricTraceKNew([3/8,7/8],[0,3/4],t^-1,q : ZigZagD := 0);
      if (GF(q)!2)^((q-1) div 4) eq -1 then
        tr3 *:= -1;
      end if;
      tr3 *:= 2;
    else
      tr3 := 0;
    end if;

    tr0 +:= q*(tr1+tr2+tr3+trivtr);

  elif family eq "L2L2" then
    tr1 := HypergeometricTrace(HypergeometricData([[1/4,3/4],[0,1/2]]),t,q);
    tr1 *:= (-1)^((q-1) div 2);

    if q mod 4 eq 1 then
      trivtr := 8;
      tr2 := HypergeometricTrace(HypergeometricData([[1/8,3/8,5/8,7/8],[0,1/4,1/2,3/4]]),t,q);
      tr2 *:= 2*(-1)^((q-1) div 4);
      if not IsSquare(GF(q)!psi) then
        tr2 *:= -1;
      end if;
    else
      trivtr := 0;
      tr2 := 0;
    end if;

    tr0 +:= q*(tr1+tr2+trivtr);

  elif family eq "L4" then
    if q mod 5 eq 1 then
      tr1 := HypergeometricTrace(HypergeometricData([[1/5,2/5,3/5,4/5],[0,1/4,1/2,3/4]]),t^-1,q);
      tr1 *:= 4;
    else
      tr1 := 0;
    end if;
        
    trivtr := 2;
    
    tr0 +:= q*(tr1+trivtr);
    
  else
    error "Family not recognized!";
  end if;

  return tr0;
end function;